#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <Unix/sys_types.h>
#include "stdio2.h"
#include "stringil.h"

extern unsigned short int w, h, bandh;

/* returns palette if color, etc. */
extern unsigned int mapmax;
extern RGBColorType map[256];

/* passed to PNG (maybe others) to generate greyscale or colors */
Boolean color;

extern char **lines, *editln;

int doPNG(FILE * fp);
void cleanPNG();
int doGIF(FILE * fp);
void cleanGIF();

/**************************************************/

/* Depends on w, h, bandh globals */
static void dobitmap(short unsigned int xofs, short unsigned int yofs,
              short unsigned int mag,
              short unsigned int dx, short unsigned int dy,
              short unsigned int dw, short unsigned int dh,
		    short unsigned int deep)
{
  // these comments pertain to rendering to a bitmap
static  unsigned char bpx[sizeof(BitmapType) + 160+1024]; // this would be passed in

#define bp ((BitmapType *)bpx)

  unsigned char *loc, *loc2;
  unsigned short int imask, bw, i, j, j0, k, bit;

  bp->width = dw;
  bp->height = 1; // dh
  bp->pixelSize = deep;
  bp->rowBytes = deep * 20; // (deep * (dw+15) / 8) & ~1, i.e. word boundary
  bp->version = 2;

  loc = (void *) bp;
  loc += sizeof(BitmapType);
  memset(loc, 0, bp->rowBytes); /* really should be bkgd, not 0 */ // * dh

  bw = 8 / deep;
  imask = 0xff << (8 - deep);
  imask &= 0xff;

  if (dw * mag + xofs >= w)
    j0 = (w / mag - xofs + bw - 1) / bw;
  else 
    j0 = (dw + bw - 1) / bw;

  for (i = 0; yofs < h && i < dh; i++, yofs += mag) {

    loc = (void *) bp;
    loc += sizeof(BitmapType);
    // loc += i * bp->rowBytes

    loc2 = lines[yofs / bandh] + w * (yofs % bandh) + xofs;

    j = j0;
    switch (deep) {
    case 8:
      if (mag == 1)
        while (j--)
          *loc++ = *loc2++;
      else
        while (j--) {
          *loc++ = *loc2;
          loc2 += mag;
        }
      break;

    case 4:
      if (mag == 1)
        while (j--) {
          k = 0xf0 & *loc2++;
          k |= *loc2++ >> 4;
          *loc++ = ~k;
      } else
        while (j--) {
          k = 0xf0 & *loc2;
          loc2 += mag;
          k |= *loc2 >> 4;
          loc2 += mag;
          *loc++ = ~k;
        }

      break;
    default:                   /* 2 or 1 */
      while (j--) {
        *loc = 0;
        bit = 0;
        do {
          k = ~*loc2 & imask;
          loc2 += mag;
          k >>= deep * bit;
          *loc |= k;
        } while (++bit < bw);
        loc++;
      }
      break;
    }

    WinDrawBitmap(bp, dx, dy + i);

  }
  if (i < dh) {
    RectangleType r = { {0, i}
    , {dh, dh - i}
    };
    WinEraseRectangle(&r, 0);   /* SHOULD BE BKGD! */
  }

}

/**************************************************/

UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
  FILE *fp;
  UInt32 deep, odeep;
  EventType event;
  int xo, yo, mag, refresh;
  char tbuf[4];

  if (cmd == sysAppLaunchCmdNormalLaunch) { /* forward pinger to zboxz */
    LocalID lid = DmFindDatabase(0, "ZBoxZ");
    SysUIAppSwitch(0, lid, sysAppLaunchCmdNormalLaunch, NULL);
    return 0;
  }

  if (cmd != sysAppLaunchCmdOpenDB || !cmdPBP || !strlen(cmdPBP))
    return 0;

  WinDrawChars(cmdPBP, StrLen(cmdPBP), 0, 145);

  fp = FileOpen(0, cmdPBP, 'DATA', 'BOXR',
                fileModeReadOnly | fileModeAnyTypeCreator, NULL);

  if (!fp)
    return 1;

  deep = 1;
  ScrDisplayMode(scrDisplayModeGetSupportedDepths, NULL, NULL, &odeep, NULL);
  ScrDisplayMode(scrDisplayModeGetSupportsColor, NULL, NULL, NULL, &color);

  if (odeep & 0x80)
    deep = 8;
  else if (odeep & 8)
    deep = 4;
  else if (odeep & 2)
    deep = 2;

  ScrDisplayMode(scrDisplayModeGet, NULL, NULL, &odeep, NULL);

  fread(tbuf, 1, 4, fp);
  rewind(fp);

  if (!strncmp(tbuf, "\211PNG", 4)) {
    if (doPNG(fp)) {
      WinDrawChars("PNG File decode did not finish", 31, 0, 0);
      SysTaskDelay(200);
      return 1;
    }
  } else if (!strncmp(tbuf, "GIF8", 4)) {
    if (doGIF(fp)) {
      cleanGIF();
      WinDrawChars("GIF File decode did not finish", 31, 0, 0);
      SysTaskDelay(200);
      return 1;
    }
  } else {
    WinDrawChars("Unrecognized file type", 22, 0, 0);
    SysTaskDelay(200);
    return 1;
  }

  if (color)
    WinPushDrawState();

  ScrDisplayMode(scrDisplayModeSet, NULL, NULL, &deep, NULL);

  if (color)
    WinPalette(winPaletteSet, 0, mapmax, map);

  xo = w / 2 - 80, yo = h / 2 - 80, mag = 0, refresh = 1;
  do {

    EvtResetAutoOffTimer();
    EvtGetEvent(&event, refresh ? 0 : -1);

    if (event.eType == nilEvent && !refresh)
      continue;

    {
      RectangleType r;
      Coord dx = xo, dy = yo, dm = mag;


      if (event.eType == keyDownEvent) {
        switch (event.data.keyDown.chr) {
        case '1'...'8':
          mag = event.data.keyDown.chr - '0';
          refresh = 1;
          break;
        case pageUpChr:
          mag++;
          refresh = 1;
          break;
        case pageDownChr:
          mag--;
          refresh = 1;
          break;
        case vchrLaunch:
	  event.eType = appStopEvent;
	  break;
        }
      }

      if (SysHandleEvent(&event))
        continue;

      if (event.eType == penDownEvent) {
        if (event.screenY > 160)
          continue;
        xo += event.screenX - 80;
        yo += event.screenY - 80;
        refresh = 1;
      }

      if (refresh) {
	refresh = 0;

        if (mag < 1)
          mag = 1;
        if (mag > 8)
          mag = 8;
        if (xo < 0)
          xo = 0;
        if (yo < 0)
          yo = 0;
        if (xo + 160 * mag > w)
          xo = w - 161 * mag + 1;
        if (yo + 160 * mag > h)
          yo = h - 161 * mag + 1;
        if (xo < 0)
          xo = 0;
        if (yo < 0)
          yo = 0;

        if (dm != mag || mag != 1) {
          dobitmap(xo, yo, mag, 0, 0, 160, 160, deep);
          continue;
        }

        dx -= xo;
        dy -= yo;

        if (dx > 0) {
          r.topLeft.x = 0;
          r.extent.x = 160 - dx;
        } else {
          r.topLeft.x = -dx;
          r.extent.x = 160 + dx;
          dx = 0;
        }

        if (dy > 0) {
          r.topLeft.y = 0;
          r.extent.y = 160 - dy;
        } else {
          r.topLeft.y = -dy;
          r.extent.y = 160 + dy;
          dy = 0;
        }

        WinCopyRectangle(NULL, NULL, &r, dx, dy, winPaint);

        /* minimal draw */

        if (dy) {
          dobitmap(xo, yo, mag, 0, 0, 160, dy, deep);
          if (dx)
            dobitmap(xo, yo + dy, mag, 0, dy, dx, 160 - dy, deep);
          if (r.topLeft.x)
            dobitmap(xo + 160 - r.topLeft.x, yo + dy, mag,
                     160 - r.topLeft.x, dy, r.topLeft.x, 160 - dy, deep);
        } else if (r.topLeft.y) {
          dobitmap(xo, yo + 160 - r.topLeft.y, mag, 0, 160 - r.topLeft.y,
                   160, r.topLeft.y, deep);
          if (dx)
            dobitmap(xo, yo, mag, 0, 0, dx, 160 - r.topLeft.y, deep);
          if (r.topLeft.x)
            dobitmap(xo + 160 - r.topLeft.x, yo, mag, 160 - r.topLeft.x, 0,
                     r.topLeft.x, 160 - r.topLeft.y, deep);
        } else if (dx)
          dobitmap(xo, yo, mag, 0, 0, dx, 160, deep);
        else if (r.topLeft.x)
          dobitmap(xo + 160 - r.topLeft.x, yo, mag, 160 - r.topLeft.x, 0,
                   r.topLeft.x, 160, deep);
        continue;

      }

    }

  } while (event.eType != appStopEvent);

  if (color)
    WinPalette(winPaletteSetToDefault, 0, 256, NULL);

  ScrDisplayMode(scrDisplayModeSet, NULL, NULL, &odeep, NULL);

  if (color)
    WinPopDrawState();

  cleanPNG();                   /* works for both */

  return 0;
}
