#define IGNORE_STDIO_STUBS
#define __string_h

#include <PalmOS.h>

#include <Unix/sys_socket.h>
#include <Unix/sys_types.h>
#include <System/SysEvtMgr.h>
#include "stringil.h"
#include "stdio2.h"

#include "SysZLib.h"


void diagprogress(RectangleType *r) {
  int i = 2;
  while ( i-- ) {
    r->topLeft.y += 8, r->topLeft.x -= 8;
    WinDrawRectangle(r, 0);
  }
}

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{

  struct prchead {
    char name[32];              //0-32
    short int attr;             //32-33
    short int vers;             //34-35
    long int cr, md, bkt;       //times 36-47
    long int mn, app, sort;     // zero 48-59 - zero for prcs.
    long int type, crea;        //60-67
    long int uidseed, nxrec;    //68-75 - uidseed rand, nxrec zero;
    short int nrecs;            //76-78
  } head;

  struct rsrcdbent {
    long int rsrc;
    short int rsid;
    long int ofst;              //80+10*(nrecs+index)
  } *rsrcent;

  struct {
    UInt16 cn;
    LocalID li;
  } *x;

  char name[36];
  LocalID lid, aiid, siid;
  int i, k;
  UInt32 asz, ssz, ofst, count, total;
  DmOpenRef db;
  void *ap;
  UInt16 cardno = 0;
  FILE *fd, *ofd;
  z_stream pgpz;
  char *ibuf, *obuf;
  Err err;
  RectangleType r;


  if (cmd == sysAppLaunchCmdSystemReset) {
    if(( lid = DmFindDatabase(cardno, "RUNZIPRUNFILE")))
      DmDeleteDatabase(0, lid);
    return 0;
  }
  else if (cmd != sysAppLaunchCmdOpenDB && cmd != sysAppLaunchCmdNormalLaunch)
    return 0;

  if (!cmdPBP)
    return 0;

  x = cmdPBP;

  DmDatabaseInfo(x->cn, x->li, name, NULL, NULL, NULL, NULL,
                 NULL, NULL, NULL, NULL, NULL, NULL);
  WinDrawChars(name, strlen(name), 10, 20);

  fd =
    FileOpen(0, name, 'DATA', 'TZIP', fileModeReadOnly | fileModeAnyTypeCreator,
             &err);

  if (err)
    return 1;

  ofd = FileOpen(0, "RUNZIPSCRATCH", 'TEMP', 'TZIP', fileModeReadWrite, NULL);

  memset(&pgpz, 0, sizeof(pgpz));
  ZLSetup;
  inflateInit2(&pgpz, -15);

#define BSIZ 2048
#define ZBSIZ (BSIZ*3)

  ibuf = MemPtrNew(BSIZ);
  obuf = MemPtrNew(ZBSIZ);

  r.topLeft.x = 0 , r.topLeft.y = 0, r.extent.y = 16;

  FileTell(fd, &total, NULL);
  while (!feof(fd)) {

    pgpz.avail_in = fread(ibuf, 1, BSIZ, fd);
    pgpz.next_in = ibuf;

    r.extent.x = 160 * pgpz.total_in / total;
    WinDrawRectangle(&r, 0);

    while (pgpz.avail_in || feof(fd)) {
      pgpz.next_out = obuf;
      pgpz.avail_out = ZBSIZ;

      k = inflate(&pgpz, !feof(fd) ? 0 : Z_FINISH);
      /* SHOULD CHECK FOR ERRORS */
      fwrite(obuf, 1, ZBSIZ - pgpz.avail_out, ofd);

      if (k == Z_BUF_ERROR && pgpz.avail_out != ZBSIZ && feof(fd))
        continue;               /* for undoc zlib */

      if (k != Z_OK)
        break;
    }
  }
  r.extent.x = 160;
  WinDrawRectangle(&r, 0);

  r.extent.y = 8;
  r.extent.x = 24;
  r.topLeft.y = 8, r.topLeft.x = 136;

  diagprogress(&r);

  MemPtrFree(obuf);
  MemPtrFree(ibuf);

  inflateEnd(&pgpz);
  ZLTeardown;

  fclose(fd);
  fd = ofd;

  diagprogress(&r);

  rewind(fd);
  FileTell(fd, &total, NULL);

  FileControl(fileOpDestructiveReadMode, fd, NULL, NULL);

  memset(&head, 0, sizeof(head));
  fread(&head, sizeof(head), 1, fd);
  count = sizeof(head);


  lid = DmFindDatabase(cardno, "RUNZIPRUNFILE");
  if (!lid) {
    DmCreateDatabase(cardno, "RUNZIPRUNFILE", 'TZIP', 'TEMP', true);
    lid = DmFindDatabase(cardno, "RUNZIPRUNFILE");
  }

  diagprogress(&r);

  i = sizeof(struct rsrcdbent) * head.nrecs;
  rsrcent = MemPtrNew(sizeof(struct rsrcdbent) + i);
  fread(rsrcent, 1, i, fd);
  count += i;
  rsrcent[head.nrecs].ofst = total;
  ofst = rsrcent[0].ofst;

  diagprogress(&r);

  db = DmOpenDatabase(cardno, lid, dmModeReadWrite);

  // clean first
  i = DmNumResources(db);
  while (i--)
    DmRemoveResource(db, i);

  diagprogress(&r);

  asz = 0;
  ssz = 0;
  if (head.app)
    asz = (head.sort ? head.sort : ofst) - head.app;
  if (head.sort)
    ssz = ofst - head.sort;

  diagprogress(&r);

#define dmfread(bufP, objSize, numObj, stream) \
FileDmRead((stream), bufP, 0, (objSize), (numObj), NULL)

  if (asz) {
    while (count < head.app)
      count +=
        fread(name, 1, head.app - count > 32 ? 32 : head.app - count, fd);
    ap = DmNewHandle(db, asz);
    count += dmfread(MemHandleLock(ap), 1, asz, fd);
    MemHandleUnlock(ap);
    aiid = MemHandleToLocalID(ap);
    DmSetDatabaseInfo(cardno, lid, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
                      &aiid, NULL, NULL, NULL);
  }

  diagprogress(&r);

  if (ssz) {
    while (count < head.sort)
      count +=
        fread(name, 1, head.sort - count > 32 ? 32 : head.sort - count, fd);
    ap = DmNewHandle(db, ssz);
    count += dmfread(MemHandleLock(ap), 1, ssz, fd);
    MemHandleUnlock(ap);
    siid = MemHandleToLocalID(ap);
    DmSetDatabaseInfo(cardno, lid, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
                      NULL, &siid, NULL, NULL);
  }

  diagprogress(&r);

  r.topLeft.y = 144, r.topLeft.x = 0, r.extent.y = 16;

  for (i = 0; i < head.nrecs; i++) {

    r.extent.x = ((i + 1) << 5) / head.nrecs * 5;
    WinDrawRectangle(&r, 0);

    while (count < rsrcent[i].ofst && !feof(fd))
      count +=
        fread(name, 1,
              rsrcent[i].ofst - count > 32 ? 32 : rsrcent[i].ofst - count, fd);
    ssz = rsrcent[i + 1].ofst - rsrcent[i].ofst;
    ap = DmNewResource(db, rsrcent[i].rsrc, rsrcent[i].rsid, ssz);
    count += dmfread(MemHandleLock(ap), 1, ssz, fd);
    MemHandleUnlock(ap);
    DmReleaseResource(ap);
  }

  fclose(fd);
  MemPtrFree(rsrcent);
  DmCloseDatabase(db);
  unlink("RUNZIPSCRATCH");

  /* At this point, RUNZIPRUNFILE is an image of all the resources in
     the original - to reconstruct the original .prc, simply set the
     name (which might conflict with the existing compressed database)
     and the rest of the header stuff - probably higher up like MakeDB
     does */

  SysAppLaunch(cardno, lid, launchFlags, sysAppLaunchCmdNormalLaunch, NULL,
               &total);

#if 1
  db = DmOpenDatabase(cardno, lid, dmModeReadWrite);

  i = DmNumResources(db);
  while (i--)
    DmRemoveResource(db, i);

  DmCloseDatabase(db);
#endif
  return 0;
}
