#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <Unix/sys_types.h>
#include "stdio2.h"
#include "stringil.h"
#include "zfaxz.h"

static const unsigned char counts[256] = {
  0x00, 0x01, 0x01, 0x02, 0x01, 0x02, 0x02, 0x03,
  0x01, 0x02, 0x02, 0x03, 0x02, 0x03, 0x03, 0x04,
  0x10, 0x11, 0x11, 0x12, 0x11, 0x12, 0x12, 0x13,
  0x11, 0x12, 0x12, 0x13, 0x12, 0x13, 0x13, 0x14,
  0x10, 0x11, 0x11, 0x12, 0x11, 0x12, 0x12, 0x13,
  0x11, 0x12, 0x12, 0x13, 0x12, 0x13, 0x13, 0x14,
  0x20, 0x21, 0x21, 0x22, 0x21, 0x22, 0x22, 0x23,
  0x21, 0x22, 0x22, 0x23, 0x22, 0x23, 0x23, 0x24,
  0x10, 0x11, 0x11, 0x12, 0x11, 0x12, 0x12, 0x13,
  0x11, 0x12, 0x12, 0x13, 0x12, 0x13, 0x13, 0x14,
  0x20, 0x21, 0x21, 0x22, 0x21, 0x22, 0x22, 0x23,
  0x21, 0x22, 0x22, 0x23, 0x22, 0x23, 0x23, 0x24,
  0x20, 0x21, 0x21, 0x22, 0x21, 0x22, 0x22, 0x23,
  0x21, 0x22, 0x22, 0x23, 0x22, 0x23, 0x23, 0x24,
  0x30, 0x31, 0x31, 0x32, 0x31, 0x32, 0x32, 0x33,
  0x31, 0x32, 0x32, 0x33, 0x32, 0x33, 0x33, 0x34,
  0x10, 0x11, 0x11, 0x12, 0x11, 0x12, 0x12, 0x13,
  0x11, 0x12, 0x12, 0x13, 0x12, 0x13, 0x13, 0x14,
  0x20, 0x21, 0x21, 0x22, 0x21, 0x22, 0x22, 0x23,
  0x21, 0x22, 0x22, 0x23, 0x22, 0x23, 0x23, 0x24,
  0x20, 0x21, 0x21, 0x22, 0x21, 0x22, 0x22, 0x23,
  0x21, 0x22, 0x22, 0x23, 0x22, 0x23, 0x23, 0x24,
  0x30, 0x31, 0x31, 0x32, 0x31, 0x32, 0x32, 0x33,
  0x31, 0x32, 0x32, 0x33, 0x32, 0x33, 0x33, 0x34,
  0x20, 0x21, 0x21, 0x22, 0x21, 0x22, 0x22, 0x23,
  0x21, 0x22, 0x22, 0x23, 0x22, 0x23, 0x23, 0x24,
  0x30, 0x31, 0x31, 0x32, 0x31, 0x32, 0x32, 0x33,
  0x31, 0x32, 0x32, 0x33, 0x32, 0x33, 0x33, 0x34,
  0x30, 0x31, 0x31, 0x32, 0x31, 0x32, 0x32, 0x33,
  0x31, 0x32, 0x32, 0x33, 0x32, 0x33, 0x33, 0x34,
  0x40, 0x41, 0x41, 0x42, 0x41, 0x42, 0x42, 0x43,
  0x41, 0x42, 0x42, 0x43, 0x42, 0x43, 0x43, 0x44
};

extern unsigned short int w, h, bandh;

/* returns palette if color, etc. */
extern unsigned int mapmax;
extern RGBColorType map[256];

/* passed to PNG (maybe others) to generate greyscale or colors */
Boolean color;

extern char **lines, *editln;

extern int doFAX(FILE * fp);
extern void cleanFAX();

static UInt32 deep;

/**************************************************/

/* Depends on w, h, bandh globals */
static void dobitmap(short unsigned int xofs, short unsigned int yofs,
                     short unsigned int dx, short unsigned int dy,
                     short unsigned int dw, short unsigned int dh)
{
  // these comments pertain to rendering to a bitmap
#define bp ((BitmapType *)bpx)
  static unsigned char bpx[sizeof(BitmapType) + 160 + 1024]; // this would be passed in
  unsigned char *loc, *loc2;
  unsigned short int i, j, k;

  bp->width = dw;
  bp->height = 1;               // dh
  bp->pixelSize = deep;
  bp->rowBytes = dw / (8 / deep);
  bp->version = 2;

  loc = (void *) bp;
  loc += sizeof(BitmapType);
  memset(loc, 0, bp->rowBytes); /* really should be bkgd, not 0 */// * dh

  if (dw + xofs >= w)
    j = (w - xofs + 7) / 8;
  else
    j = (dw + 7) / 8;

  yofs &= ~(deep - 1);

  for (i = 0; yofs < h && i < dh; i++, yofs += deep) {
    loc = (void *) bp;
    loc += sizeof(BitmapType);

    loc2 = lines[yofs >> 8] + 216 * (yofs & 0xff) + xofs / 8;
    k = j * deep;

    switch (deep) {
    case 1:
      memcpy(loc, loc2, k);
      break;
    case 4:
      while (k--) {
        *loc++ =
          counts[loc2[0]] + counts[loc2[216] & 0xbf] +
          counts[loc2[432] & 0xfd] + counts[loc2[648]];
        loc2++;
      }
      break;
    }
    //    memcpy(loc, loc2, j);
    WinDrawBitmap(bp, dx, dy + i);
  }

  if (i < dh) {
    RectangleType r = { {0, i}
    , {dh, dh - i}
    };
    WinEraseRectangle(&r, 0);   /* SHOULD BE BKGD! */
  }
}

/**************************************************/

void faxdisp(char *cp)
{
  FILE *fp;
  UInt32 odeep,sdeep;
  EventType event;
  long int xo, yo;
  int refresh;
  RectangleType r;
  Coord dx, dy, margin;

  WinDrawChars(cp, strlen(cp), 0, 145);

  fp = FileOpen(0, cp, 'DATA', 'BOXR',
                fileModeReadOnly | fileModeAnyTypeCreator, NULL);
  if (!fp)
    return;

  doFAX(fp);

  ScrDisplayMode(scrDisplayModeGetSupportedDepths, NULL, NULL, &sdeep, NULL);

  deep = 1;
  if (sdeep & 8)
    deep = 4;

  ScrDisplayMode(scrDisplayModeGet, NULL, NULL, &odeep, NULL);
  ScrDisplayMode(scrDisplayModeSet, NULL, NULL, &deep, NULL);

  xo = yo = dx = dy = 0;
  refresh = 0;
  margin = 160 * deep;
  dobitmap(0, 0, 0, 0, 160, 160);


  do {

    EvtResetAutoOffTimer();
    EvtGetEvent(&event, refresh ? 0 : -1);

    if (event.eType == nilEvent && !refresh)
      continue;

    if (event.eType == keyDownEvent)
      switch (event.data.keyDown.chr) {
      case pageUpChr:
        if (deep == 1)
          break;
        deep = 1;
        ScrDisplayMode(scrDisplayModeSet, NULL, NULL, &deep, NULL);
        margin = 160 * deep;
        dobitmap(xo, yo, 0, 0, 160, 160);
        continue;
      case pageDownChr:
        if (deep == 4 || !(sdeep & 8))
          break;
        deep = 4;
        ScrDisplayMode(scrDisplayModeSet, NULL, NULL, &deep, NULL);
        margin = 160 * deep;
        dobitmap(xo, yo, 0, 0, 160, 160);
        continue;
      case vchrLaunch:
	event.eType = appStopEvent;
	break;
      }

    if (SysHandleEvent(&event))
      continue;

    dx = xo, dy = yo;
    if (event.eType == penDownEvent || event.eType == penMoveEvent) {
      if (event.screenY > 160)
        continue;
      xo = event.screenX;
      yo = event.screenY;
      refresh = 1;
    }

    if (refresh) {
      refresh = 0;

      xo -= 4;
      if (xo > 152)
        xo = w - margin;
      else
        xo = xo * (w - margin) / 152;
      if (xo < 0)
        xo = 0;

      yo -= 4;
      if (yo > 152)
        yo = h - margin;
      else
        yo = yo * (h - margin) / 152;
      if (yo < 0)
        yo = 0;

      xo &= ~(8 / deep - 1);
      yo &= ~(deep - 1);

      // This needs to be seriously modified for 4bpp.

      dx -= xo;
      dy -= yo;

      if (deep > 1 || dy > 159 || dx > 159 || dx < -159 || dy < -159) {
        dobitmap(xo, yo, 0, 0, 160, 160);
        continue;
      }

      if (dx > 0) {
        r.topLeft.x = 0;
        r.extent.x = 160 - dx;
      } else {
        r.topLeft.x = -dx;
        r.extent.x = 160 + dx;
        dx = 0;
      }

      if (dy > 0) {
        r.topLeft.y = 0;
        r.extent.y = 160 - dy;
      } else {
        r.topLeft.y = -dy;
        r.extent.y = 160 + dy;
        dy = 0;
      }

      WinCopyRectangle(NULL, NULL, &r, dx, dy, winPaint);

      /* minimal draw */

      if (dy) {
        dobitmap(xo, yo, 0, 0, 160, dy);
        if (dx)
          dobitmap(xo, yo + dy, 0, dy, dx, 160 - dy);
        if (r.topLeft.x)
          dobitmap(xo + 160 - r.topLeft.x, yo + dy,
                   160 - r.topLeft.x, dy, r.topLeft.x, 160 - dy);
      } else if (r.topLeft.y) {
        dobitmap(xo, yo + 160 - r.topLeft.y, 0, 160 - r.topLeft.y,
                 160, r.topLeft.y);
        if (dx)
          dobitmap(xo, yo, 0, 0, dx, 160 - r.topLeft.y);
        if (r.topLeft.x)
          dobitmap(xo + 160 - r.topLeft.x, yo, 160 - r.topLeft.x, 0,
                   r.topLeft.x, 160 - r.topLeft.y);
      } else if (dx)
        dobitmap(xo, yo, 0, 0, dx, 160);
      else if (r.topLeft.x)
        dobitmap(xo + 160 - r.topLeft.x, yo, 160 - r.topLeft.x, 0,
                 r.topLeft.x, 160);
      continue;

    }
  } while (event.eType != appStopEvent);

  ScrDisplayMode(scrDisplayModeSet, NULL, NULL, &odeep, NULL);
  cleanFAX();
  return;
}
